// *************************************************************************
//    Unify.C
// *************************************************************************

#include "Unify.h"

#include "Config.h"
#include "Trail.h"
#include "CoralError.h"
#include "Globals.h"
#include "TupleIterator.h"
#include "HashSimpleRelation.h"
#include "Execute.h"
#include "Debug.h"

//#define   SHOW_UNIFY_RESULT    1

#ifdef    SHOW_UNIFY_RESULT
#    define    RET_SUCCEED    do { \
        left.printon(&cout); \
        cout << " vs "; \
        right.printon(&cout); \
        cout << " - succeeded. (@ " << __LINE__ << ")"; \
        cout << " (" << left.expr << ")\n"; \
        return(COR_U_SUCCEED); } while(0)

#    define    RET_FAIL    do { \
        left.printon(&cout); \
        cout << " vs "; \
        right.printon(&cout); \
        cout << " - failed. (@ " << __LINE__ << ")\n"; \
        return(COR_U_FAIL); } while(0)

#else
#    define    RET_SUCCEED   return(COR_U_SUCCEED) 
#    define    RET_FAIL      return(COR_U_FAIL)
#endif

// *************************************************************************

#define DO_OCCUR_CHECK 0

static Cor_Term left_term, right_term;

// *************************************************************************
//  What we do to please purify.
void clean_up_static_terms()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    left_term.expr.init();
    right_term.expr.init();
}

// *************************************************************************
//
// unify_binding:
// This is used to bind a variable to a term.  The variable must be free.
// The term may actually be a variable interpreted in a bindenv.
// If so, it is dereferenced completely. This is actually essential
// so that a variable is never bound to itself -- that would be a silly
// (and wrong) thing to do.
//
/*******************************
%     if (DO_OCCUR_CHECK) {
%     // WARNING:: Incomplete implementation
%         // Simple check - don't bind variable to itself, but return success.
%         if (!(base_term->expr->is_var()
%               && (base_term->var_num() == var)
%               && base_term->bindenv == bindenv) )
%         return(COR_U_SUCCEED);
%     // Fill in the rest of the occur check code here ....
%     }
********************************/

// *************************************************************************
inline int Cor_inline_unify_binding(Cor_BindEnv *bindenv, Variable var,
                                    Cor_Term term)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

#ifdef DO_TRACE
    if (Cor_exEnv->dbg_unify) {
        *(Cor_exEnv->trace_strm)
            << "[unify_binding(bindenv=" <<  bindenv
            << ", var#=" << var << ", arg=" << &(term.expr) << "=";
        term.expr.printon(Cor_exEnv->trace_strm);
        *(Cor_exEnv->trace_strm) << ", bindenv=" << term.bindenv << ")]\n";
    }
#endif

    Cor_full_dereference_term(term);

    if ( ! ((term.bindenv == bindenv)
           && term.expr.is_var()
           && (term.expr.var_num() == var)) ) {

        ASSERT(bindenv);
        bindenv->bind(var, term);
        Cor_CurTrail->PushTrail(bindenv, var);
    }
    /* else:  don't bind variable to itself -- simply return success. */

#ifdef DO_TRACE
    if (Cor_exEnv->dbg_unify)
        *(Cor_exEnv->trace_strm) << "unify_binding succeeded.\n";
#endif
    return(COR_U_SUCCEED);
}

// *************************************************************************
int Cor_unify_binding(Cor_BindEnv *bindenv, Variable var, Cor_Term term)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return Cor_inline_unify_binding(bindenv, var, term);
}

// *************************************************************************
int Cor_subsumes_binding(Cor_BindEnv *bindenv, Variable var, Cor_Term term,
                         int can_bind)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef DO_TRACE
    if (Cor_exEnv->dbg_unify) {
        *(Cor_exEnv->trace_strm)
                << "[subsumes_binding(bindenv=" << bindenv
                << ", var=" << var
                << ", arg=" << &(term.expr) << "=" << term.expr
                << ", bindenv=" << term.bindenv << ")]\n";
    }
#endif

    ASSERT((bindenv != NULL) && (bindenv->lookup(var).expr.is_empty()));

    Cor_Term base_term = term;
    Cor_full_dereference_term(base_term);

    if (base_term.expr.is_var()
            && (base_term.varnum() == var)
            && (base_term.bindenv == bindenv))
        return(COR_U_SUCCEED);

    if (!can_bind)
        return(COR_U_FAIL);

    /*
    if (DO_OCCUR_CHECK) {
        // WARNING: Incomplete implementation
    }
    */

    bindenv->bind(var, base_term);
    Cor_CurTrail->PushTrail(bindenv, var);

#ifdef DO_TRACE
    if (Cor_exEnv->dbg_unify)
        *(Cor_exEnv->trace_strm) << 
                                         "subsumes_binding succeeded.\n";
#endif
    return(COR_U_SUCCEED);
}

// *************************************************************************
// Unify the terms 'left' and 'right'.
// Return COR_U_SUCCEED or COR_U_FAIL.
// 
// WARNING:  COR_DONTCARE is dangerous:  make sure there is no underscore
// within a structure.  If you unify f(_) with f(a), the result on the lhs
// will be f(_), not f(a).  Thus the result may depend on which side you
// choose!!  This is shady.  In particular, the result of unifying
// g(X,X,X) with g(f(_),f(a),Y) is order dependent!!
// *************************************************************************
int Cor_unify_args(Cor_Term left, Cor_Term right)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    CArgKind left_Kind, right_Kind;

#ifdef DO_PROFILE
    if (Cor_T_Stack->count > 0)
        (Cor_T_Stack->top()->moduledata
         ->module_info->Cor_GlobalProfileInfo.unifies)++;
    else
        (Cor_GlobalProfileInfo.unifies)++;
#endif

#ifdef DO_TRACE
    if (Cor_exEnv->dbg_unify) {
        *(Cor_exEnv->trace_strm) << "[unify_args(left=" << &left.expr << "=";
        left.printon(Cor_exEnv->trace_strm);
        *(Cor_exEnv->trace_strm) << ", right=" << &right.expr << "=";
        right.printon(Cor_exEnv->trace_strm);
        *(Cor_exEnv->trace_strm) << ", left_bindings="
            << left.bindenv << ", right_bindings="
            << right.bindenv << ")]\n";
    }
#endif

    Cor_full_dereference_term(left);

    for (;;) {

/*
        // SPF
        cout << "Unify:  " << left << " vs " << right;
        cout << " (" << left.expr << " vs " << right.expr << ")";
        cout << endl;
*/

        if ((left.bindenv == right.bindenv)
                    && left.expr.equals(&(right.expr)))
            RET_SUCCEED;

        left_Kind  = left.Kind();
        right_Kind = right.Kind();

        switch (left_Kind) {

        case VAR_ARG:
            if (right_Kind != DONT_CARE_ARG) {
#ifdef    SHOW_UNIFY_RESULT
            {
                int k = Cor_inline_unify_binding(left.bindenv, left.varnum(),
                                                right);
                if (k==COR_U_SUCCEED)
                    RET_SUCCEED;
                else
                    RET_FAIL;
            }
#else
                return Cor_inline_unify_binding(left.bindenv, left.varnum(),
                                                right);
#endif
              // Left was fully dereferenced earlier
            } else
                RET_SUCCEED;

        case INT_ARG:
        case DOUBLE_ARG:
            switch (right_Kind) {

            case ADD_OP_ARG: 
            case SUB_OP_ARG: 
            case MULT_OP_ARG: 
            case DIV_OP_ARG: 
            case IDIV_OP_ARG: 
            case MOD_OP_ARG: 
            case LSHFT_OP_ARG: 
            case RSHFT_OP_ARG: 
            case BT_AND_OP_ARG: 
            case BT_OR_OP_ARG: 
            case BT_NOT_OP_ARG: 
            case BT_XOR_OP_ARG: 
                Cor_arith_dereference_term(right);
                // Now fall through to number check

            case INT_ARG:
            case DOUBLE_ARG:
                if (left.expr.compare_num(right.expr))
                    RET_FAIL;
                RET_SUCCEED;

            case VAR_ARG:
                if (right.bindenv->not_bound(right.expr.var_num()))
#ifdef    SHOW_UNIFY_RESULT
                {
                    int k = Cor_inline_unify_binding(
                                    right.bindenv, right.varnum(), left);
                    if (k == COR_U_SUCCEED)
                        RET_SUCCEED;
                    else
                        RET_FAIL;
                }
#else
                    return Cor_inline_unify_binding(
                                    right.bindenv, right.varnum(), left);
#endif
                else {
                    Cor_full_dereference_term(right);
                    continue;
                }

            case DONT_CARE_ARG:
                RET_SUCCEED;

            default:
                RET_FAIL;
            }

        case SYMBOL_ARG:
        case TINY_SYM_ARG:
            switch (right_Kind) {

            case SYMBOL_ARG:
            case TINY_SYM_ARG:
                if (left.expr.equals(right.expr))
                    RET_SUCCEED;
                else
                    RET_FAIL;

            case VAR_ARG:
                if (right.bindenv->not_bound(right.expr.var_num()))
#ifdef    SHOW_UNIFY_RESULT
                {
                    int k = Cor_inline_unify_binding(
                                    right.bindenv, right.varnum(), left);
                    if (k==COR_U_SUCCEED)
                        RET_SUCCEED;
                    else
                        RET_FAIL;
                }
#else
                    return Cor_inline_unify_binding(
                                    right.bindenv, right.varnum(), left);
#endif
                else {
                    Cor_full_dereference_term(right);
                    continue;
                }

            case DONT_CARE_ARG:
                RET_SUCCEED;

            default:
                RET_FAIL;
            }

        case FUNCT_ARG:
            switch (right_Kind) {

            case VAR_ARG:
                if (right.bindenv->not_bound(right.expr.var_num()))
#ifdef    SHOW_UNIFY_RESULT
                {
                    int k= Cor_inline_unify_binding(
                                    right.bindenv, right.varnum(), left);
                    if (k==COR_U_SUCCEED)
                        RET_SUCCEED;
                    else
                        RET_FAIL;
                }
#else
                    return Cor_inline_unify_binding(
                                    right.bindenv, right.varnum(), left);
#endif
                else {
                    Cor_full_dereference_term(right);
                    continue;
                }

            case FUNCT_ARG: {

                if ( (left.expr.arity() != right.expr.arity())
                        || ( !left.expr.nm_equals(right.expr.functor()) ))
                    RET_FAIL;

       // NOTE:
       // In the following code, after checking if a term is
       // fully ground, we compute the hash value.  isConstant
       // stores VarHashValue if it finds that there is a variable
       // in the term.  If the term is in fact
       // a constant, hash stores the hash value.
       //
       // Hence, so long as this pair is always used together
       // no term is traversed more than twice by these functions.
       // If we call only isConstant, ground terms can be traversed
       // many times by these functions.
       // If we call just hash, nonground terms can be traversed
       // many times by these functions.

                HashVal   left_hash, right_hash;
                int       left_const;
                Cor_CArg *left_func  = &(left.expr);
                Cor_CArg *right_func = &(right.expr);

                if ( Cor_isGroundHashValue(left_hash = left_func->hash())) {
                    left_const = 1;
                } else {
                    left_hash  = left_func->hash(left.bindenv);
                    left_const = Cor_isGroundHashValue(left_hash);
                }

                if (left_const) {
                  if (Cor_isGroundHashValue(right_hash = right_func->hash())){

                    if (left_hash != right_hash)
                        RET_FAIL;

                  } else {
                    right_hash  = right_func->hash(right.bindenv);
                    if (Cor_isGroundHashValue(right_hash)
                            && (left_hash != right_hash))
                        RET_FAIL;
                  }
                }

#ifdef    SHOW_UNIFY_RESULT
                {
                    int k = Cor_unify_arg_lists(left_func->args(),
                        right_func->args(), left.bindenv, right.bindenv);
                    if (k==COR_U_SUCCEED)
                        RET_SUCCEED;
                    else
                        RET_FAIL;
                }
#else
                return(Cor_unify_arg_lists(left_func->args(),
                        right_func->args(), left.bindenv, right.bindenv));
#endif
            }                // case FUNCT_ARG


            case SUM_GRPING_ARG:
            case PROD_GRPING_ARG:
            case CNT_GRPING_ARG:
            case AVG_GRPING_ARG:
            case STDDEV_GRPING_ARG:
            case SUM_DGRPING_ARG:
            case PROD_DGRPING_ARG:
            case CNT_DGRPING_ARG:
            case AVG_DGRPING_ARG:
            case MIN_GRPING_ARG:
            case MAX_GRPING_ARG:
            case SET_GRPING_ARG:
            case ANY_GRPING_ARG:
            case NO_AGG_GRPING_ARG:
                CORAL_error(COR_INTERNAL_ERROR, "Grouping in unification.",
                            "unify_args()");

            default:
                RET_FAIL;

            }        // left=FUNCT_ARG, switch(right->Kind())

        case NIL_ARG:
            switch (right_Kind) {
       
            case VAR_ARG: 
                if (right.bindenv->not_bound(right.expr.var_num()))
#ifdef    SHOW_UNIFY_RESULT
                {
                    int k= Cor_inline_unify_binding(
                                    right.bindenv, right.varnum(), left);
                    if (k==COR_U_SUCCEED)
                        RET_SUCCEED;
                    else
                        RET_FAIL;
                }
#else
                    return Cor_inline_unify_binding(
                                    right.bindenv, right.varnum(), left);
#endif
                else {
                    Cor_full_dereference_term(right);
                    continue;
                }

            case NIL_ARG: 
                RET_SUCCEED;

            default:
                RET_FAIL;
            }

        case LIST_ARG:
            switch (right_Kind) {

            case VAR_ARG:
                if (right.bindenv->not_bound(right.expr.var_num()))
#ifdef    SHOW_UNIFY_RESULT
                {
                    int k= Cor_inline_unify_binding(
                                    right.bindenv, right.varnum(), left);
                    if (k==COR_U_SUCCEED)
                        RET_SUCCEED;
                    else
                        RET_FAIL;
                }
#else
                    return Cor_inline_unify_binding(
                                    right.bindenv, right.varnum(), left);
#endif
                else {
                    Cor_full_dereference_term(right);
                    continue;
                }

            case LIST_ARG: {

                HashVal   left_hash, right_hash;
                int       left_const;
                Cor_CArg *left_list  = &(left.expr);
                Cor_CArg *right_list = &(right.expr);

                if ( Cor_isGroundHashValue(left_hash = left_list->hash())) {
                    left_const = 1;
                } else {
                    left_hash  = left_list->hash(left.bindenv);
                    left_const = Cor_isGroundHashValue(left_hash);
                }

                if (left_const) {
                  if (Cor_isGroundHashValue(right_hash = right_list->hash())){

                    if (left_hash != right_hash)
                        RET_FAIL;

                  } else {
                    right_hash  = right_list->hash(right.bindenv);
                    if (Cor_isGroundHashValue(right_hash)
                            && (left_hash != right_hash))
                        RET_FAIL;
                  }
                }

                  // First unify car()
                if ( Cor_unify_arg( *(left_list->car()), *(right_list->car()),
                         left.bindenv, right.bindenv) == COR_U_FAIL)
                    RET_FAIL;

                Cor_CArg left_cdr, right_cdr;
                left_list->get_cdr(left_cdr);
                right_list->get_cdr(right_cdr);

                  // Then unify cdr() 
#ifdef    SHOW_UNIFY_RESULT
                {
                int k= Cor_unify_arg(left_cdr, right_cdr,
                       left.bindenv, right.bindenv);
                    if (k==COR_U_SUCCEED)
                        RET_SUCCEED;
                    else
                        RET_FAIL;
                }
#else
                return(Cor_unify_arg(left_cdr, right_cdr,
                       left.bindenv, right.bindenv));
#endif

            }           // case LIST_ARG

            default:
                RET_FAIL;

            }        // left=LIST_ARG, switch(right->Kind())


        case DONT_CARE_ARG:
            RET_SUCCEED;

        case REL_ARG: {

            switch (right_Kind) {

            case VAR_ARG:{
                    if (right.bindenv->not_bound(right.expr.var_num()))
#ifdef    SHOW_UNIFY_RESULT
                    {
                        int k= Cor_unify_binding(
                                right.bindenv, right.varnum(), left);
                        if (k==COR_U_SUCCEED)
                            RET_SUCCEED;
                        else
                            RET_FAIL;
                    }
#else
                        return Cor_unify_binding(
                                right.bindenv, right.varnum(), left);
#endif
                    else {
                        Cor_full_dereference_term(right);
#ifdef    SHOW_UNIFY_RESULT
                    {
                        int k= Cor_unify_args(right, left);
                        if (k==COR_U_SUCCEED)
                            RET_SUCCEED;
                        else
                            RET_FAIL;
                    }
#else
                        return Cor_unify_args(right, left);
#endif
                    }
                }

            case REL_ARG: {

                    Cor_HashSimpleRelation *left_rel
                            = (Cor_HashSimpleRelation *) left.expr.rel();
                    Cor_HashSimpleRelation *right_rel
                            = (Cor_HashSimpleRelation *) right.expr.rel();

                    if (left_rel->equals(right_rel) == 1)
                        RET_SUCCEED;

                    RET_FAIL;
                }

            case SUM_GRPING_ARG:
            case PROD_GRPING_ARG:
            case CNT_GRPING_ARG:
            case AVG_GRPING_ARG:
            case STDDEV_GRPING_ARG:
            case SUM_DGRPING_ARG:
            case PROD_DGRPING_ARG:
            case CNT_DGRPING_ARG:
            case AVG_DGRPING_ARG:
            case MIN_GRPING_ARG:
            case MAX_GRPING_ARG:
            case SET_GRPING_ARG:
            case ANY_GRPING_ARG:
            case NO_AGG_GRPING_ARG:
                CORAL_error(COR_INTERNAL_ERROR, "Grouping in unification.",
                            "unify_args()");

            default:
                RET_FAIL;
            }
        }

        case SUM_GRPING_ARG:
        case PROD_GRPING_ARG:
        case CNT_GRPING_ARG:
        case AVG_GRPING_ARG:
        case STDDEV_GRPING_ARG:
        case SUM_DGRPING_ARG:
        case PROD_DGRPING_ARG:
        case CNT_DGRPING_ARG:
        case AVG_DGRPING_ARG:
        case MIN_GRPING_ARG:
        case MAX_GRPING_ARG:
        case SET_GRPING_ARG:
        case ANY_GRPING_ARG:
        case NO_AGG_GRPING_ARG:
            CORAL_error(COR_INTERNAL_ERROR, "Grouping in unification.",
                        "unify_args()");
            RET_FAIL;

        default:
#ifdef    DEBUG
            cerr << "No case for " << left.expr.kind_str()
                 << " in " << __FUNCTION__ << endl; 
#endif
            *(Cor_exEnv->err_strm) << 
                  "Error! unify: default case yet to be implemented!\n";
            RET_FAIL;
        }        // switch (left->Kind())

    }            // for (;;)

    cerr << "Reached end of unify_args().\n";
    RET_FAIL;
#undef    RET_SUCCEED
#undef    RET_FAIL
}              // unify_args

// *************************************************************************
int Cor_unify_arg(Cor_CArg& left, Cor_CArg& right,
               Cor_BindEnv *left_bindings, Cor_BindEnv *right_bindings)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Term lTerm(left,  left_bindings);
    Cor_Term rTerm(right, right_bindings);

    return(Cor_unify_args(lTerm,rTerm)); 
}

// *************************************************************************
int Cor_unify_arg_lists(Cor_ArgList& left, Cor_ArgList& right,
               Cor_BindEnv *left_bindings, Cor_BindEnv *right_bindings)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef DO_TRACE
    if (Cor_exEnv->dbg_unify)
        *(Cor_exEnv->trace_strm) << "unify_arg_lists called:\n";
#endif

    int i;

    for (i = 0; i < left.count(); i++) {
        left_term.expr = left[i];
        right_term.expr = right[i];
        left_term.bindenv = left_bindings;
        right_term.bindenv = right_bindings;
        if (Cor_unify_args(left_term, right_term) == COR_U_FAIL)
            return(COR_U_FAIL);
    }
#ifdef DO_TRACE
    if (Cor_exEnv->dbg_unify)
        *(Cor_exEnv->trace_strm) << "unify_arg_lists succeeded.\n";
#endif

    return(COR_U_SUCCEED);
}

// *************************************************************************

extern int Cor_subsumes_nested_arg_lists(Cor_ArgList& left, Cor_ArgList& right,
    Cor_BindEnv *left_bindings, Cor_BindEnv *right_bindings, int can_bind);

#define CAN_BIND 1

// *************************************************************************
/*
 * The following fn. checks if left subsumes right
 * 
 * WARNING:  The code  assumes that left and right do not share any
 * variables! If they do, right would have to be copied
 * (recursively) so that it does not share any subterm
 * with left.
 * 
 * ASSUMPTIONS:
 * 1) We assume that right and left don't share any variables.
 * 2) right.bindenv can be NULL - this is fine since right
 *    cannot be bound in any case.  Some checks for
 *    null bindenv are required as a result.
 * 3) If the flag canbind is not set, then subsumption checks for
 *    equality - no variable binding  is allowed.
 *    The equality is strict in that no variable renaming
 *    is allowed.
 * 4) If right.bindenv is not null, we need to know when a
 *    left variable is bound to a right variable.
 */

int Cor_subsumes_args(Cor_Term left, Cor_Term right, int can_bind)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef DO_TRACE
    if (Cor_exEnv->dbg_subsumes) {
        *(Cor_exEnv->trace_strm) << "[subsumes_args(left=" << &left.expr << "=";
        left.expr.printon(Cor_exEnv->trace_strm);
        *(Cor_exEnv->trace_strm) << ", right=" << &right.expr << "=";
        right.expr.printon(Cor_exEnv->trace_strm);
        *(Cor_exEnv->trace_strm)
            << ", left_bindings=" << left.bindenv
            << ", right_bindings=" << right.bindenv << ")]\n";
    }
#endif

    if ((left.bindenv == right.bindenv)
                && (left.expr == right.expr))
            return(COR_U_SUCCEED);

    switch (left.Kind()) {

    case DONT_CARE_ARG:
        return(COR_U_SUCCEED);

    case VAR_ARG:
        {
                // Can happen is left was bound to a
                // right variable and was dereferenced
            if (left.bindenv == NULL) 
                return(COR_U_FAIL);
            if (left.bindenv->not_bound(left.expr.var_num()))
                return Cor_subsumes_binding(left.bindenv, left.varnum(), right,
                                            can_bind);
#if 1
     /* WARNING:  Assumes that the left term has a single level bindenv.
      * The condition is satisfied provided it is a fact, since
      * our fact representation has only single level bindenvs
      * MAKE SURE that this assumption is valid in other uses!!
      */
            else {
                /* Bug Fix:  the can_bind argument was never changed before.
                 * It should be changed when dereferencing crosses over
                 * into the right hand side. */
                Cor_Term left1 = left.dereference_var();
                if (left1.bindenv == left.bindenv) {
                    /* Still in left -- can bind left variable */
                    return Cor_subsumes_args(left1, right, can_bind);
                } else {
                      // Crossed over to right -- cannot bind left variable
                    return Cor_subsumes_args(left1, right, 0);
                }
            }
#else
            else
                return Cor_subsumes_args(left.dereference_var(), right,
                                         right.bindenv ? 0 : can_bind);
            // if right.bindenv is not null, left.bindenv
            // must have been empty, so dereferencing left
            // makes us cross over into right.  Hence set can_bind
            // to 0;
#endif
        }

    case ADD_OP_ARG:
    case SUB_OP_ARG:
    case MULT_OP_ARG:
    case DIV_OP_ARG:
    case IDIV_OP_ARG:
    case MOD_OP_ARG:
    case LSHFT_OP_ARG:
    case RSHFT_OP_ARG:
    case BT_AND_OP_ARG:
    case BT_OR_OP_ARG:
    case BT_NOT_OP_ARG:
    case BT_XOR_OP_ARG:
        Cor_arith_dereference_term(left);
        // Fall through to test

    case INT_ARG:
    case DOUBLE_ARG:
    case SYMBOL_ARG:
    case TINY_SYM_ARG:

        switch (right.Kind()) {

        case DONT_CARE_ARG:
            return(COR_U_FAIL);

        case VAR_ARG:{
                if ((right.bindenv == NULL)
                        || right.bindenv->not_bound(right.expr.var_num()))
                    return(COR_U_FAIL);
                return(Cor_subsumes_args(left,
                        right.dereference_var(), can_bind));
            }

        case ADD_OP_ARG:
        case SUB_OP_ARG:
        case MULT_OP_ARG:
        case DIV_OP_ARG:
        case IDIV_OP_ARG:
        case MOD_OP_ARG:
        case LSHFT_OP_ARG:
        case RSHFT_OP_ARG:
        case BT_AND_OP_ARG:
        case BT_OR_OP_ARG:
        case BT_NOT_OP_ARG:
        case BT_XOR_OP_ARG:
            Cor_arith_dereference_term(right);

        case INT_ARG:
        case DOUBLE_ARG:
            Cor_full_dereference_term(right);
            if (! left.expr.is_num()
                    || (left.expr.compare_num(right.expr)))
                return(COR_U_FAIL);
            return(COR_U_SUCCEED);

        case SYMBOL_ARG:
        case TINY_SYM_ARG:
            if (right.expr.equals(left.expr))
                return(COR_U_SUCCEED);
            else
                return(COR_U_FAIL);

        case FUNCT_ARG:
        case REL_ARG:

        case NIL_ARG:
        case LIST_ARG:
            return(COR_U_FAIL);

        case SUM_GRPING_ARG:
        case PROD_GRPING_ARG:
        case CNT_GRPING_ARG:
        case AVG_GRPING_ARG:
        case STDDEV_GRPING_ARG:
        case SUM_DGRPING_ARG:
        case PROD_DGRPING_ARG:
        case CNT_DGRPING_ARG:
        case AVG_DGRPING_ARG:
        case MIN_GRPING_ARG:
        case MAX_GRPING_ARG:
        case SET_GRPING_ARG:
        case ANY_GRPING_ARG:
        case NO_AGG_GRPING_ARG:
            CORAL_error(COR_INTERNAL_ERROR, "Grouping in subsumption check.",
                        "subsumes_args()");
        default:
            return(COR_U_FAIL);
        }

    case REL_ARG:

        switch (right.Kind()) {

        case DONT_CARE_ARG:
            return(COR_U_FAIL);

        case VAR_ARG:{
                if ((right.bindenv == NULL)
                        || right.bindenv->not_bound(right.expr.var_num()))
                    return(COR_U_FAIL);
                return(Cor_subsumes_args(left,
                            right.dereference_var(), can_bind));
            }

        case ADD_OP_ARG: 
        case SUB_OP_ARG: 
        case MULT_OP_ARG: 
        case DIV_OP_ARG: 
        case IDIV_OP_ARG: 
        case MOD_OP_ARG: 
        case LSHFT_OP_ARG: 
        case RSHFT_OP_ARG: 
        case BT_AND_OP_ARG: 
        case BT_OR_OP_ARG: 
        case BT_NOT_OP_ARG: 
        case BT_XOR_OP_ARG: 

        case INT_ARG:
        case DOUBLE_ARG:
        case SYMBOL_ARG:
        case TINY_SYM_ARG:
        case FUNCT_ARG:
        case NIL_ARG:
        case LIST_ARG:
            return(COR_U_FAIL);

        case SUM_GRPING_ARG:
        case PROD_GRPING_ARG:
        case CNT_GRPING_ARG:
        case AVG_GRPING_ARG:
        case STDDEV_GRPING_ARG:
        case SUM_DGRPING_ARG:
        case PROD_DGRPING_ARG:
        case CNT_DGRPING_ARG:
        case AVG_DGRPING_ARG:
        case MIN_GRPING_ARG:
        case MAX_GRPING_ARG:
        case SET_GRPING_ARG:
        case ANY_GRPING_ARG:
        case NO_AGG_GRPING_ARG:
            CORAL_error(COR_INTERNAL_ERROR, "Grouping in subsumption check.",
                        "subsumes_args()");

        case REL_ARG:{

                Cor_Relation *left_rel = left.expr.rel();
                Cor_Relation *right_rel = right.expr.rel();

                if (left_rel->equals(right_rel) == 1)
                    return(COR_U_SUCCEED);
                return(COR_U_FAIL);
            }

        default:
            return(COR_U_FAIL);
        }

    case FUNCT_ARG:

        switch (right.Kind()) {

        case DONT_CARE_ARG:
            return(COR_U_FAIL);

        case VAR_ARG:{
                if ((right.bindenv == NULL)
                        || right.bindenv->not_bound(right.expr.var_num()))
                    return(COR_U_FAIL);
                return(Cor_subsumes_args(left,
                                right.dereference_var(), can_bind));
                // can_bind might start out bound, but not set by subsumes
            }

        case INT_ARG:
        case DOUBLE_ARG:
        case ADD_OP_ARG: 
        case SUB_OP_ARG: 
        case MULT_OP_ARG: 
        case DIV_OP_ARG: 
        case IDIV_OP_ARG: 
        case MOD_OP_ARG: 
        case LSHFT_OP_ARG: 
        case RSHFT_OP_ARG: 
        case BT_AND_OP_ARG: 
        case BT_OR_OP_ARG: 
        case BT_NOT_OP_ARG: 
        case BT_XOR_OP_ARG: 
        case SYMBOL_ARG:
        case TINY_SYM_ARG:
        case REL_ARG:
            return(COR_U_FAIL);

        case FUNCT_ARG:{
            if (! left.expr.equals(right.expr))
                return(COR_U_FAIL);

            Cor_CArg *left_func  = &(left.expr);
            Cor_CArg *right_func = &(right.expr);

            if (left_func->isConstant()) {
                HashVal left_hash = left_func->hash(left.bindenv);
                if (Cor_isGroundHashValue(left_hash)
                        && right_func->isConstant()) {
                    HashVal right_hash = right_func->hash(right.bindenv);
                    if (Cor_isGroundHashValue(right_hash)) {
                        if (left_hash != right_hash)
                            return(COR_U_FAIL);
                        else
                            return(COR_U_SUCCEED);
                    }
                }
            }
            return(Cor_subsumes_nested_arg_lists(
                              left_func->args(), right_func->args(),
                              left.bindenv, right.bindenv, can_bind));

        }                // case FUNCT_ARG

        case SUM_GRPING_ARG:
        case PROD_GRPING_ARG:
        case CNT_GRPING_ARG:
        case AVG_GRPING_ARG:
        case STDDEV_GRPING_ARG:
        case SUM_DGRPING_ARG:
        case PROD_DGRPING_ARG:
        case CNT_DGRPING_ARG:
        case AVG_DGRPING_ARG:
        case MIN_GRPING_ARG:
        case MAX_GRPING_ARG:
        case SET_GRPING_ARG:
        case ANY_GRPING_ARG:
        case NO_AGG_GRPING_ARG:
            CORAL_error(COR_INTERNAL_ERROR, "Grouping in subsumption check.",
                        "subsumes_args()");
        default:
            return(COR_U_FAIL);
        }

    case NIL_ARG:

        switch (right.Kind()) {

        case DONT_CARE_ARG:
            return(COR_U_FAIL);

        case VAR_ARG: {
                if ((right.bindenv == NULL)
                        || right.bindenv->not_bound(right.expr.var_num()))
                    return(COR_U_FAIL);
                return(Cor_subsumes_args(left,
                                right.dereference_var(), can_bind));
                // can_bind might start out bound, but not set by subsumes
            }

        case INT_ARG:
        case DOUBLE_ARG:
        case ADD_OP_ARG: 
        case SUB_OP_ARG: 
        case MULT_OP_ARG: 
        case DIV_OP_ARG: 
        case IDIV_OP_ARG: 
        case MOD_OP_ARG: 
        case LSHFT_OP_ARG: 
        case RSHFT_OP_ARG: 
        case BT_AND_OP_ARG: 
        case BT_OR_OP_ARG: 
        case BT_NOT_OP_ARG: 
        case BT_XOR_OP_ARG: 
        case SYMBOL_ARG:
        case TINY_SYM_ARG:
        case REL_ARG:
        case FUNCT_ARG:
        case LIST_ARG:
            return(COR_U_FAIL);

        case NIL_ARG:
            return(COR_U_SUCCEED);

        case SUM_GRPING_ARG:
        case PROD_GRPING_ARG:
        case CNT_GRPING_ARG:
        case AVG_GRPING_ARG:
        case STDDEV_GRPING_ARG:
        case SUM_DGRPING_ARG:
        case PROD_DGRPING_ARG:
        case CNT_DGRPING_ARG:
        case AVG_DGRPING_ARG:
        case MIN_GRPING_ARG:
        case MAX_GRPING_ARG:
        case SET_GRPING_ARG:
        case ANY_GRPING_ARG:
        case NO_AGG_GRPING_ARG:
            CORAL_error(COR_INTERNAL_ERROR, "Grouping in subsumption check.",
                        "subsumes_args()");
        default:
            return(COR_U_FAIL);
        }                // case NIL_ARG

    case LIST_ARG:
        switch (right.Kind()) {

        case DONT_CARE_ARG:
            return(COR_U_FAIL);

        case VAR_ARG: {
                if ((right.bindenv == NULL)
                        || right.bindenv->not_bound(right.expr.var_num()))
                    return(COR_U_FAIL);
                return(Cor_subsumes_args(left,
                                right.dereference_var(), can_bind));
                // can_bind might start out bound, but not set by subsumes
            }

        case INT_ARG:
        case DOUBLE_ARG:
        case ADD_OP_ARG: 
        case SUB_OP_ARG: 
        case MULT_OP_ARG: 
        case DIV_OP_ARG: 
        case IDIV_OP_ARG: 
        case MOD_OP_ARG: 
        case LSHFT_OP_ARG: 
        case RSHFT_OP_ARG: 
        case BT_AND_OP_ARG: 
        case BT_OR_OP_ARG: 
        case BT_NOT_OP_ARG: 
        case BT_XOR_OP_ARG: 
        case SYMBOL_ARG:
        case TINY_SYM_ARG:
        case REL_ARG:
        case FUNCT_ARG:
            return(COR_U_FAIL);

        case LIST_ARG: {

                HashVal   left_hash, right_hash;
                int       left_const;
                Cor_CArg *left_list  = &(left.expr);
                Cor_CArg *right_list = &(right.expr);

                if ( Cor_isGroundHashValue(left_hash = left_list->hash())) {
                    left_const = 1;
                } else {
                    left_hash  = left_list->hash(left.bindenv);
                    left_const = Cor_isGroundHashValue(left_hash);
                }

                if (left_const) {

                  if (Cor_isGroundHashValue(right_hash = right_list->hash())){

                    if (left_hash != right_hash)
                        return(COR_U_FAIL);

                  } else {
                    right_hash  = right_list->hash(right.bindenv);
                    if (Cor_isGroundHashValue(right_hash)
                            && (left_hash != right_hash))
                        return(COR_U_FAIL);
                  }

                  if (left.expr.equals(right.expr))
                      return(COR_U_SUCCEED);
                }

            {
                Cor_Term lTerm(left_list->car(),  left.bindenv);
                Cor_Term rTerm(right_list->car(), right.bindenv);

                if (Cor_subsumes_args(lTerm, rTerm, can_bind) == COR_U_FAIL)
                    return(COR_U_FAIL);

                left_list->get_cdr(lTerm.expr);
                right_list->get_cdr(rTerm.expr);

                return(Cor_subsumes_args(lTerm, rTerm, can_bind));
            }
        }                // case LIST_ARG

        case SUM_GRPING_ARG:
        case PROD_GRPING_ARG:
        case CNT_GRPING_ARG:
        case AVG_GRPING_ARG:
        case STDDEV_GRPING_ARG:
        case SUM_DGRPING_ARG:
        case PROD_DGRPING_ARG:
        case CNT_DGRPING_ARG:
        case AVG_DGRPING_ARG:
        case MIN_GRPING_ARG:
        case MAX_GRPING_ARG:
        case SET_GRPING_ARG:
        case ANY_GRPING_ARG:
        case NO_AGG_GRPING_ARG:
            CORAL_error(COR_INTERNAL_ERROR, "Grouping in subsumption check.",
                        "subsumes_args()");
        default:
            return(COR_U_FAIL);
        }

    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case CNT_GRPING_ARG:
    case AVG_GRPING_ARG:
    case STDDEV_GRPING_ARG:
    case SUM_DGRPING_ARG:
    case PROD_DGRPING_ARG:
    case CNT_DGRPING_ARG:
    case AVG_DGRPING_ARG:
    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case SET_GRPING_ARG:
    case ANY_GRPING_ARG:
    case NO_AGG_GRPING_ARG:
        CORAL_error(COR_INTERNAL_ERROR, "Grouping in subsumption check.",
                    "subsumes_args()");

    case AGG_RESULT_ARG:
    case CMPLX_AGG_RESULT_ARG:
        TELL;
        cerr << "Not yet implemented.\n";
        break;

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Kind is: " << left.expr.kind_str() << endl;
#endif
        *(Cor_exEnv->err_strm) << 
               "Error! subsumes: default case yet to be implemented!\n";
        break;

    }

    return(COR_U_FAIL);
}

// *************************************************************************
int Cor_subsumes_nested_arg_lists(Cor_ArgList& left, Cor_ArgList& right,
    Cor_BindEnv *left_bindings, Cor_BindEnv *right_bindings, int can_bind)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef DO_TRACE
    if (Cor_exEnv->dbg_subsumes) {
        *(Cor_exEnv->trace_strm)
            << "[subsumes_nested_arg_lists(left&=" << &left;
        *(Cor_exEnv->trace_strm) << ", right&=" << &right;
        *(Cor_exEnv->trace_strm)
            << ", left_bindings=" << left_bindings
            << ", right_bindings=" << right_bindings << ")]\n";
    }
#endif


    left_term.bindenv = left_bindings;
    right_term.bindenv = right_bindings;

    int i = left.count();

    if (right.count() != i)
        return(COR_U_FAIL);

    for (; --i >= 0;) {
        left_term.expr = left[i];
        right_term.expr = right[i];
        if (Cor_subsumes_args(left_term, right_term, can_bind) == COR_U_FAIL)
            return(COR_U_FAIL);
    }
    return(COR_U_SUCCEED);
}

// *************************************************************************
int Cor_subsumes_arg_lists(Cor_ArgList &left, Cor_ArgList &right,
               Cor_BindEnv *left_bindings, Cor_BindEnv *right_bindings)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef DO_PROFILE
    if (Cor_T_Stack->count > 0)
        (Cor_T_Stack->top()->moduledata->module_info->
                Cor_GlobalProfileInfo.Cor_subsumes)++;
    else
        (Cor_GlobalProfileInfo.Cor_subsumes)++;
#endif

#ifdef DO_TRACE
    if (Cor_exEnv->dbg_subsumes) {
        *(Cor_exEnv->trace_strm)
            << "[subsumes_arg_lists(left&=" << left;
        *(Cor_exEnv->trace_strm) << ", right&=" << right;
        *(Cor_exEnv->trace_strm)
            << ", left_bindings=" << left_bindings
            << ", right_bindings=" << right_bindings << ")]\n";

        *(Cor_exEnv->trace_strm) << "Left: ";
        left_bindings->printon(Cor_exEnv->trace_strm);
        *(Cor_exEnv->trace_strm) << "\nRight: ";
        right_bindings->printon(Cor_exEnv->trace_strm);
    }
#endif


    /********  ERROR -- TO BE FIXED
    // WARNING::  ReturnUnify optimizations needed here!!!
    .. fix subsumes to assume that one of the arguments is in
    ..   a normal form where there are no cross bindenv bindings.
    ..   we can then use this fact to avoid checks for all_free etc.
    ..   Perhaps we can pass in the bindenv that is in normal form as an
    ..   argument when called from subsumes_tuple, so that all_free is
    ..   used for all current used.
    *******************************/

    /* ERROR:  By commenting out the following code, we make an assumption
     * that left bindings never cross into the right.
     * MAKE SURE that this assumption is valid (esp. in ord-search)
     */
    if (!( (right_bindings == NULL) 
          || ((left_bindings != NULL) && (left_bindings->all_free()) )
          || left.isConstant() )) {

        /*
        *(Cor_exEnv->err_strm)
            << "System Error: subsumes_arg_lists works only "
            << "when either right_bindings is null,\n\t"
            << " or left_bindings is all-free.\n";
        */
        return(COR_U_FAIL);
    }

    return Cor_subsumes_nested_arg_lists(left, right, left_bindings,
                                         right_bindings, CAN_BIND);
}

// *************************************************************************
int Cor_subsumes(Cor_BindEnv *env1, Cor_ArgList& arg2, Cor_BindEnv *env2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return (Cor_subsumes_arg_lists(arg2, arg2, env1, env2));
}

// *************************************************************************
/*
 * Checks if tuple subsumes newtuple, and prints a message if display is
 * set.  If tuple does not have a bindenv it creates one of the required
 * size.
 */

int Cor_subsumes_tuple(Cor_Tuple *tuple, Cor_Tuple *newtuple, int display)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_StackMark unify_mark;
    int stat;

#ifdef    DEBUG
    //cout << "Subsumption check: " << *tuple << " vs " << *newtuple << endl;
#endif

    static Cor_BindEnv NilEnv(0);

    stat = Cor_subsumes_arg_lists(tuple->args(), newtuple->args(),
                                      &NilEnv, NULL);
    unify_mark.pop_to();

    if (stat && display) {
        *(Cor_exEnv->trace_strm) << ") - subsumed by (";
        tuple->printon(Cor_exEnv->trace_strm);
        *(Cor_exEnv->trace_strm) << ")]\n";
    }

    return(stat);
}

// *************************************************************************
int Cor_subsumes_value(CArgKind aggkind, Cor_CArg val1, Cor_CArg val2,
                       Cor_BindEnv *env1, Cor_BindEnv *env2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Returns 1 if val1 (new value) subsumes val2 (old value),
    // -1 if val2 (old value) subsumes val1 (new value) and
    // returns 0 if the values are equal.


    Cor_StackMark stackmark;
    Cor_Term term1(val1, env1);
    Cor_Term term2(val2, env2);
    Cor_full_dereference_term(term1);
    Cor_full_dereference_term(term2);

    // First dispose of "any" selection
    if (aggkind == ANY_GRPING_ARG) {

        int stat = Cor_unify_args(term1, term2);
        stackmark.pop_to();
        if (stat == COR_U_SUCCEED)
            return(0);
        else
            return(-1);        // new value is subsumed if don't unify

    }

    if (! term1.expr.is_num()
            || ! term2.expr.is_num() ) {
        *(Cor_exEnv->err_strm) << 
              "Aggregate Selection: aggregated argument not a number\n";
        stackmark.pop_to();
        return(0);
    }

    int stat = (term1.expr.compare_num(term2.expr));
    // stat is pos if term1.expr is greater, neg if lesser, and 0 if equal

    if (aggkind == MIN_GRPING_ARG) {
        stat = -stat;
    } else if (aggkind != MAX_GRPING_ARG) {
        TELL;
        CORAL_error(COR_BAD_ANNO, "Bad aggregation type", "subsumes_value");
        stat = 0;
    }

    stackmark.pop_to();
    return(stat);
}

// *************************************************************************
//    Checks if each of tuple1 and tuple2 subsumes the other.

int Cor_equal_tuples(Cor_Tuple * tuple1, Cor_Tuple * tuple2, int display)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(Cor_subsumes_tuple(tuple1, tuple2, display) &&
            Cor_subsumes_tuple(tuple2, tuple1, display));
}

// *************************************************************************
int Cor_unify_literal(Cor_TupleIterator& iterator, Cor_Tuple *tuple,
        int display)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_StackMark unify_mark;

    if (Cor_unify_arg_lists(iterator.arg_list, tuple->args(),
                            iterator.bindenv, iterator.tuple_env)) {

        // Remove this line ?? ___ EFFICIENCY1
        iterator.reset_no_match();
#ifdef DO_TRACE
        if (display) {
            *(Cor_exEnv->trace_strm) << 
                    "<subclass>Relation: get_next_tuple: ";
            iterator.relation->name.print(NULL, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << "(";
            tuple->printon(Cor_exEnv->trace_strm);
        }
#endif

        return(1);
    }
#ifdef DO_PROFILE
    if (Cor_T_Stack->count > 0)
        (Cor_T_Stack->top()->moduledata->module_info->
         Cor_GlobalProfileInfo.non_unifying_tuples)++;
    else
        (Cor_GlobalProfileInfo.non_unifying_tuples)++;
#endif

    unify_mark.pop_to();
    return(0);
}

//#define    UNIFY_VERBOSE    1

// *************************************************************************
// Assign the term 'dbtrm' to 'vartrm'.
// Return COR_U_SUCCEED or COR_U_FAIL.
// *************************************************************************
int Cor_Assign_args(Cor_Term& vartrm, Cor_Term& dbtrm)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int vn;
    ASSERT(vartrm.expr.is_var());

    vn = vartrm.expr.var_num();
    ASSERT(vartrm.bindenv->not_bound(vn));

    vartrm.bindenv->bind(vn, dbtrm);
    Cor_CurTrail->PushTrail(vartrm.bindenv, vn);

    return(COR_U_SUCCEED);
}

// *************************************************************************
// Assign the arg 'dbarg' to 'vararg'.
// Return COR_U_SUCCEED or COR_U_FAIL.
// *************************************************************************
int Cor_Assign_args(Cor_CArg *vararg, Cor_BindEnv *env, Cor_CArg *dbarg)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int vn;
    ASSERT(vararg->is_var());

    vn = vararg->var_num();
    ASSERT(env->not_bound(vn));

    env->bind(vn, dbarg);
    Cor_CurTrail->PushTrail(env,vn);

    return(COR_U_SUCCEED);
}

// *************************************************************************
// Equal Check the terms 'vartrm' and 'dbtrm'.
// Return COR_U_SUCCEED or COR_U_FAIL.
// *************************************************************************
int Cor_EqlChk_args(Cor_Term& vartrm, Cor_Term& dbtrm)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
      // vartrm might be a bound variable
    Cor_full_dereference_term(vartrm);

    ASSERT(! vartrm.expr.is_var());

    if (vartrm.expr.equals(&(dbtrm.expr))) {
        return(COR_U_SUCCEED);
    }

    return(COR_U_FAIL);
}

// *************************************************************************
// Equal Check the terms 'vartrm' and 'dbarg'.
// Return COR_U_SUCCEED or COR_U_FAIL.
// *************************************************************************
int Cor_EqlChk_args(Cor_Term& vartrm, Cor_CArg *dbarg)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
      // vartrm might be a bound variable
    Cor_full_dereference_term(vartrm);

    ASSERT(! vartrm.expr.is_var());

    if (vartrm.expr.equals(dbarg)) {
        return(COR_U_SUCCEED);
    }

    return(COR_U_FAIL);
}

int Cor_EqlChk_args(Cor_CArg *var_arg, Cor_BindEnv *var_env,
            Cor_CArg *dbarg)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
      // var_arg might be a bound variable
    if (! var_arg->is_empty()) {

        while (var_env && var_arg->is_var()) {

            var_arg = var_env->full_lookup(var_arg->var_num(), &var_env);

            if (var_arg->is_empty())
                break;
        }

          // Damn, need a Term now
        if (var_arg->is_expr()) {
            Cor_Term term(var_arg,var_env);

            Cor_arith_dereference_term(term);

            if (term.expr.equals(dbarg))
                return(COR_U_SUCCEED);
            else
                return(COR_U_FAIL);
        }
    }

    ASSERT(! var_arg->is_var());

    if (var_arg->equals(dbarg))
        return(COR_U_SUCCEED);

    return(COR_U_FAIL);
}

// *************************************************************************
int Cor_Match_Arg( Cor_CArg *var_arg, Cor_BindEnv *var_env,
            Cor_CArg *db_arg )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int vn;
    
    switch(var_arg->Kind()) {

    case VAR_ARG:
        if (var_env->not_bound(vn = var_arg->var_num())) {
            var_env->bind(vn,db_arg);
            Cor_CurTrail->PushTrail(var_env, vn);
            return(COR_U_SUCCEED);
        } else {
            return( Cor_Match_Arg(var_env->carg_lookup(vn), var_env, db_arg) );
        }

    case FUNCT_ARG: {

            int arity = var_arg->arity();

            if (!db_arg->is_functor()
                    || (arity != db_arg->arity())
                    || !var_arg->cfunctor()->equals(db_arg->cfunctor()) )
                return(COR_U_FAIL);

            Cor_ArgList &my_var_al(var_arg->args());
            Cor_ArgList &my_db_al(db_arg->args());

            int j;
            for (j=arity-1; j >= 0; j--) {
                Cor_CArg *my_var_arg, *my_db_arg; 

                my_var_arg = &(my_var_al[j]);
                my_db_arg  = &(my_db_al[j]);

                if (Cor_Match_Arg(my_var_arg, var_env, my_db_arg)
                        == COR_U_FAIL) {
                    return(COR_U_FAIL);
                }
            }
            return(COR_U_SUCCEED);
        }

    case LIST_ARG: {
            if (!db_arg->is_cons())
                return(COR_U_FAIL);

            if (Cor_Match_Arg(var_arg->car(), var_env,
                        db_arg->car()) == COR_U_FAIL)
                return(COR_U_FAIL);

            Cor_CArg var_cdr, db_cdr;
            var_arg->get_cdr(var_cdr);
            db_arg->get_cdr(db_cdr);
            return(Cor_Match_Arg(&var_cdr, var_env, &db_cdr));
        }

    default:
        // all other Constant types
        if (var_arg->equals(db_arg))
            return(COR_U_SUCCEED);
        else
            return(COR_U_FAIL);
    }
}

// *************************************************************************
// Match the terms 'vartrm' and 'dbtrm'.
// Return COR_U_SUCCEED or COR_U_FAIL.
// *************************************************************************
int Cor_Match_args(Cor_Term& vartrm, Cor_Term& dbtrm)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(Cor_Match_Arg( &(vartrm.expr), vartrm.bindenv, &(dbtrm.expr) ));
}

// *************************************************************************
