// *************************************************************************
//    Database.C
// *************************************************************************

#include <stdio.h>
#include "Database.h"

#include "SymTable.h"
#include "CoralError.h"
#include "Globals.h"
#include "RelationMisc.h"
#include "Interface.h"
#include "Debug.h"

// *************************************************************************
Cor_DatabaseStruct *Cor_lookupDB(Cor_CArg& name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    if (Cor_DatabaseTable == NULL)
        Cor_DatabaseTable = new Cor_SymTable();
    
    return((Cor_DatabaseStruct *) Cor_DatabaseTable->lookup(name));
}

// *************************************************************************
Cor_DatabaseStruct *Cor_createDB(Cor_CArg& name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char err_msg[35];
    Cor_DatabaseStruct *db;

    if ((db = Cor_lookupDB(name))) {
        sprintf(err_msg, "%10s.. already exists", name.symbol());
        err_msg[34] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "createDB");
        return(NULL);
    }
   
    db = new Cor_DatabaseStruct(name);
    Cor_DatabaseTable->insert(name, db);

    db->RelationTable = new Cor_SymTable();
    return(db); 
}

// *************************************************************************
// used by exit_coral to remove all DBs

void Cor_clearDatabases()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    Cor_SymCursor *cursor  = Cor_DatabaseTable->OpenScan();

    do {

        Cor_DatabaseStruct *db_strt = (Cor_DatabaseStruct *) cursor->data();

        if (db_strt && (! db_strt->name.equals(COR_BUILTINS)) )
            delete db_strt;

    } while(cursor->ScanNext());

    delete cursor;
}

// *************************************************************************
// Used by init_coral to add the default DB to the DatabaseTable

int Cor_addDBStruct(Cor_DatabaseStruct *db)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    if (db) {

        if (Cor_lookupDB(db->name)) {
            return(0);
        }

        Cor_DatabaseTable->insert_nodup(db->name, db);
        return(1);

    } else
        return(0);
}

// *************************************************************************
int Cor_destroyDB(Cor_CArg& name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char err_msg[35];

    if (name.equals(Cor_CurDB->name)) {
        CORAL_error(COR_MULTIPLE_DB_ERR, "cannot destroy current DB",
            "destroyDB");
        return(0);
    }

    if (name.equals(COR_BUILTINS)) {
        *(Cor_exEnv->err_strm) << "Cannot destroy the builtin database.\n";
        return(0);
    }

    Cor_DatabaseStruct *db;

    if (! (db = (Cor_DatabaseStruct *)
            Cor_DatabaseTable->delete_entry(name))) {
        sprintf(err_msg, "%10s does not exist", name.symbol());
        err_msg[34] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "destroyDB");
        return(0);
    }

    delete db;
    return(1);
}

// *************************************************************************
int Cor_currentDB(Cor_CArg& name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char err_msg[35];
    Cor_DatabaseStruct *db;

    if (name.equals(COR_BUILTINS)) {
        *(Cor_exEnv->err_strm)
            << "Cannot make the builtin database the current database\n";
        return(0);
    }

    if (!(db = Cor_lookupDB(name))) {
        sprintf(err_msg, "%10s does not exist", name.symbol());
        err_msg[34] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "currentDB");
        return(0);
    }

    Cor_CurDB = db;
    return 1;
}

// *************************************************************************
int Cor_renameDB(Cor_CArg& dbname, Cor_CArg& newName)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char err_msg[35];
    Cor_DatabaseStruct *db;

    if (dbname.equals(COR_BUILTINS)) {
        *(Cor_exEnv->err_strm) << "Cannot rename the builtin database\n";
        return(0);
    }

    if (!(db = Cor_lookupDB(dbname))) {
        sprintf(err_msg, "%10s does not exist", dbname.symbol());
        err_msg[34] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "renameDB");
        return(0);
    }

    if (Cor_lookupDB(newName)) {
        sprintf(err_msg, "%10s already exists", newName.symbol());
        err_msg[34] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "renameDB");
        return(0);
    }

    Cor_DatabaseTable->delete_entry(dbname);
    Cor_DatabaseTable->insert(newName, db);
    db->name = newName;

    return(1);
}

// *************************************************************************
int Cor_renameRelation(Cor_CArg& relname, Cor_CArg& newName)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Relation *oldrel;

    oldrel = (Cor_Relation *) Cor_CurDB->RelationTable->lookup(relname);

    if (! oldrel) {
        *(Cor_exEnv->err_strm) << relname.symbol() << " does not exist";
        return(0);
    }

    if (Cor_CurDB->RelationTable->lookup(newName)) {
        *(Cor_exEnv->err_strm) << newName.symbol() << " already exists";
        return(0);
    }

    Cor_CurDB->RelationTable->delete_entry(relname);
    Cor_CurDB->RelationTable->insert(newName, oldrel);
    oldrel->name = newName;

    return(1);
}

// *************************************************************************
int Cor_saveDB(Cor_CArg& dbname, ostream *outf)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char err_msg[35];
    Cor_DatabaseStruct *db;

    if (dbname.is_empty())
        db = Cor_CurDB;
    else if (dbname.equals(COR_BUILTINS)) {
        *(Cor_exEnv->err_strm) << "Cannot save the builtin database.\n";
        return(0);
    }
    else if (!(db = Cor_lookupDB(dbname))) {
        sprintf(err_msg, "%10s does not exist", dbname.symbol());
        err_msg[34] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "saveDB");
        return(0);
    }

    Cor_SymCursor *cursor  = db->RelationTable->OpenScan();

    do {

        Cor_Relation *rel = (Cor_Relation *) cursor->data();
        if (rel)
            rel->print_facts(outf);

    } while(cursor->ScanNext());

    delete cursor;
    return(1);
}

// *************************************************************************
int Cor_copyToDB(Cor_CArg *srcrel, Cor_CArg& dbname, Cor_CArg *dstrel)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_DatabaseStruct *db;
    char err_msg[100];

    if (!(db = Cor_lookupDB(dbname))) {
        sprintf(err_msg, "%10s DB does not exist", dbname.symbol());
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "copyToDB");
        return(0);
    }

    Cor_CArg src_nme(srcrel->functor());

    Cor_Relation *src = (Cor_Relation *)
            db->RelationTable->lookup(src_nme);

    if (! src) {
        sprintf(err_msg,"rel %10s does not exist in DB %s",
                src_nme.symbol(), dbname.symbol());
        CORAL_error(COR_BAD_REL_NAME, err_msg, "copyToDB");
        return(0);
    }

    Cor_CArg dst_nme(dstrel->functor());

    Cor_Relation *dst = (Cor_Relation *)
            Cor_CurDB->RelationTable->lookup(dst_nme);

    if (dst) {
        sprintf(err_msg,"rel %10s ..already exists in current DB",
                    dstrel->functor());
        CORAL_error(COR_BAD_REL_NAME, err_msg, "copyToDB");
        return(0);
    }

    dst = Cor_make_relation(*dstrel, dstrel->args().arity());

    Cor_Tuple *tuple;
    FOR_EACH_TUPLE(tuple, src) {
        dst->insert_new(tuple);
    } END_EACH_TUPLE(tuple)

    return(1);
}

// *************************************************************************
int Cor_addToDB(Cor_CArg *srcrel, Cor_CArg& dbname, Cor_CArg *dstrel)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_DatabaseStruct *db;
    char err_msg[100];

    if (!(db = Cor_lookupDB(dbname))) {
        sprintf(err_msg, "%10s DB does not exist", dbname.symbol());
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "addToDB");
        return(0);
    }

    Cor_CArg src_nme(srcrel->functor());

    Cor_Relation *src = (Cor_Relation *)
            db->RelationTable->lookup(src_nme);

    if (! src) {
        sprintf(err_msg,"rel %10s does not exist in DB %s",
                    srcrel->functor(), dbname.symbol());
        CORAL_error(COR_BAD_REL_NAME, err_msg, "addToDB");
        return(0);
    }

    Cor_CArg dst_nme(dstrel->functor());

    Cor_Relation *dst = (Cor_Relation *)
            Cor_CurDB->RelationTable->lookup(dst_nme);

    if (! dst) {
        sprintf(err_msg,"rel %10s does not exist in current DB",
                    dstrel->functor());
        CORAL_error(COR_BAD_REL_NAME, err_msg, "addToDB");
        return(0);
    }

    Cor_Tuple *tuple;
    FOR_EACH_TUPLE(tuple, src) {
        dst->insert_new(tuple);
    } END_EACH_TUPLE(tuple)

    return(1);
}

// *************************************************************************
int Cor_removeRelFromDB(Cor_CArg *rel, Cor_CArg& dbname)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_DatabaseStruct *db;

    if (dbname.equals(Cor_CurDB->name))
        db = Cor_CurDB;
    else if (!(db = Cor_lookupDB(dbname))) {
        *(Cor_exEnv->err_strm) << "Error: db " << dbname
                               << " does not exist.\n";
        return(0);
    }

    Cor_CArg relnme = Cor_CArg(rel->functor());

    Cor_Relation *rm_rel = (Cor_Relation *)
            db->RelationTable->delete_entry(relnme);

    if (! rm_rel) {

        *(Cor_exEnv->err_strm) << "Error: rel " << relnme
                               << " does not exist.\n";
        return(0);

    } else {

        if (! --(rm_rel->ref_count)) {
            rm_rel->empty_relation(1);
            delete rm_rel;
        }
    }

    return(1);
}

// *************************************************************************
int Cor_clearRelInDB(Cor_CArg *rel, Cor_CArg& dbname)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char err_msg[45];

    Cor_DatabaseStruct *db;
    if (dbname.equals(Cor_CurDB->name))
        db = Cor_CurDB;

    else if (!(db = Cor_lookupDB(dbname))) {
        sprintf(err_msg, "db %10s does not exist", dbname.symbol());
        err_msg[44] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "clearRelInDB");
        return(0);
    }

    Cor_CArg rel_arg = Cor_CArg(rel->functor());
    Cor_Relation *clr_rel = (Cor_Relation *)
            db->RelationTable->lookup(rel_arg);

    if (! clr_rel) {
        sprintf(err_msg,"rel %10s does not exist", rel->functor());
        err_msg[44] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "clearRelInDB");
        return(0);
    } else
        clr_rel->empty_relation(1);

    return(1);
}

// *************************************************************************
static void InsertRel(Cor_DatabaseStruct *db, Cor_CArg& n, int,
            Cor_Relation *r)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!db->RelationTable)
        db->RelationTable = new Cor_SymTable();

        // this does not check for duplicates
    db->RelationTable->insert(n, (Pointer) r);
}


// *************************************************************************
int Cor_openRelInCurDB(Cor_CArg *rel, Cor_CArg& dbname)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char err_msg[35];
    Cor_DatabaseStruct *db;

    if (dbname.equals(COR_BUILTINS)) {
        *(Cor_exEnv->err_strm)
                << "Access already exists to builtin relations.\n";
        return(0);
    }
    else if (!(db = Cor_lookupDB(dbname))) {
        sprintf(err_msg, "%10s does not exist", dbname.symbol());
        err_msg[34] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "openRelInCurDB");
        return(0);
    }

    Cor_CArg rel_nme(rel->functor());

    Cor_Relation *opn_rel = (Cor_Relation *)
            db->RelationTable->lookup(rel_nme);

    if (opn_rel) {
        InsertRel(Cor_CurDB, rel_nme, rel->args().arity(), opn_rel);
        (void) (opn_rel->ref_count)++;
        return(1);
    }
    
    CORAL_error(COR_REL_NOT_FOUND, rel->functor(),
            "OpenRelInCurDB");
    return(0);
}

// *************************************************************************
int Cor_listEveryDB()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    Cor_SymCursor *cursor  = Cor_DatabaseTable->OpenScan();

    do {

        *(Cor_exEnv->err_strm) << cursor->arg().symbol() << endl;

    } while(cursor->ScanNext());

    delete cursor;
    return(1);
}

// *************************************************************************
int Cor_listRelsInDB(Cor_CArg& dbname, ostream *outfile )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char err_msg[35];
    Cor_DatabaseStruct *db;

    if (!(db = Cor_lookupDB(dbname))) {
        sprintf(err_msg, "%10s does not exist", dbname.symbol());
        err_msg[34] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "currentDB");
        return(0);
    }

    Cor_SymCursor *cursor  = db->RelationTable->OpenScan();

    if (cursor) {
        do {
            Cor_Relation *rel = (Cor_Relation *) cursor->data();
            if (rel)
                rel->print_name(outfile);

        } while(cursor->ScanNext());
    }

    delete cursor;
    return(1);
}

// *************************************************************************
