// *************************************************************************
//    SymTable.C
// *************************************************************************


#include "SymTable.h"
#include "Globals.h"
#include "Debug.h"

// *************************************************************************
Cor_SymTabElement::Cor_SymTabElement()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    data = NULL;
    _next = NULL; 
}

// *************************************************************************
Cor_SymTabElement::Cor_SymTabElement(Cor_CArg& ste_arg, Pointer ste_data)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    arg = ste_arg;

    data = ste_data;
    _next = NULL;
}

// *************************************************************************
Cor_SymTabElement::~Cor_SymTabElement()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
}

// *************************************************************************
Cor_SymCursor::Cor_SymCursor(Cor_SymTable *_table)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    table = _table;

        // The next attempt will get the first item from index 0
    index = 0;
    current = table->_symtab[0];

    if (! current)
        ScanNext();
}

// *************************************************************************
Cor_SymCursor::~Cor_SymCursor()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
}

// *************************************************************************
int Cor_SymCursor::ScanNext()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (index < 0)
        return(0);

    if (current){

        current = current->_next;
        if (current)
            return(1);
    }

    int limit = table->size();
    for (index++; index < limit; index++) {
        if ((current = table->_symtab[index]))
            return(1);
    }

    index = -1;
    return(0);
}

// *************************************************************************
Cor_SymTable::Cor_SymTable(int size)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (size == 0)
        size = COR_DEFLT_SYMTAB_SIZE;

    _size = size;

    _num_elements = 0;

        // Rehash larger when more than 2 elements (on average)
    _element_limit = _size + _size;

    _symtab = new Cor_SymTabElement*[_size];

    ASSERT(_symtab != NULL);

    for (int j=0; j < _size; j++)
        _symtab[j] = NULL;
}

// *************************************************************************
Cor_SymTable::~Cor_SymTable()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    clear();
    delete [] _symtab;
}

// *************************************************************************
// Warning, this will insert duplicates, data cannot be NULL
int  Cor_SymTable::insert(Cor_CArg& arg, Pointer data)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(this != NULL);
    ASSERT(data != NULL);

    HashVal hval = arg.symhash();

    Cor_SymTabElement *element = new Cor_SymTabElement(arg,data);

    int indx = hval % _size;
    //*cerr_strm << "In insert ("
    //    << arg.symbol()
    //    << ")  hval: " << hval
    //    << "  _size: " << _size
    //    << "   indx: " << indx << endl;

    element->set_next(_symtab[indx]);
    _symtab[indx] = element;

    if (++_num_elements > _element_limit)
        rehash_bigger();

    return(0);
}

// *************************************************************************
// This one checks for duplicates before inserting, data cannot be NULL
Pointer Cor_SymTable::insert_nodup(Cor_CArg& arg, Pointer data)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(this != NULL);
    ASSERT(data != NULL);

    HashVal hval = arg.symhash();

    Cor_SymTabElement *element = new Cor_SymTabElement(arg,data);

    int indx = hval % _size;

    Pointer was_dup = delete_entry(arg);

    element->set_next(_symtab[indx]);
    _symtab[indx] = element;

    if (++_num_elements > _element_limit)
        rehash_bigger();

    return(was_dup);
}

// *************************************************************************
    // Returns Pointer to data if delete found its target
Pointer  Cor_SymTable::delete_entry(Cor_CArg& arg)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(this != NULL);

    Pointer was_del = NULL;

    HashVal hval = arg.symhash();
    Cor_SymTabElement *before, *now;

    int indx = hval % _size;

    for (before = NULL, now = _symtab[indx]; now;
            before = now, now = now->get_next() ) {
        
        if (arg.equals(now->arg)) {

            was_del = now->data;

            if (before)
                before->set_next(now->get_next());
            else
                _symtab[indx] = now->get_next();

            break;
        }
    }

    if (now) {
        delete now;
        _num_elements--;
    }
    return(was_del);
}

// *************************************************************************
// Returns NULL if arg is not in table, otherwise the data.
Pointer Cor_SymTable::lookup(Cor_CArg& arg)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(this != NULL);

    if (arg.is_empty())
        return(NULL);

    HashVal hval = arg.symhash();
    Cor_SymTabElement *now;

    int indx = hval % _size;

    for (now = _symtab[indx]; now; now = now->get_next() ) {
       
        //*cerr_strm << "In lookup(indx:" << indx
        //    << "), comparing " << arg.symbol()
        //    << " with " now->arg.symbol() << endl;
        if (arg.equals(now->arg))
            return(now->data);
    }

    return(NULL);
}

// *************************************************************************
// Delete all items in the table
void Cor_SymTable::clear()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(this != NULL);

    int j;
    Cor_SymTabElement *element, *prev = NULL;

    for (j=0; j < _size; j++) {
        
        for (element = _symtab[j]; element; element = element->get_next()){
            delete prev;
            prev = element;
        }
    }

        // Only one left
    delete prev;

        // Clear out the table in case its being re-used
    for (j=0; j < _size; j++)
        _symtab[j] = NULL;
}

// *************************************************************************
// Double the size of the hash table and rehash all elements
void Cor_SymTable::rehash_bigger()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(this != NULL);

    HashVal hval;
    int j, indx;

    int old_size = _size;
    Cor_SymTabElement **old_symtab = _symtab;

    _size = _size + _size;
    _element_limit = _size + _size;
    _symtab = new Cor_SymTabElement*[_size];

    for (j=0; j < _size; j++)
        _symtab[j] = NULL;


        // Rehash the old items into the new table
    for (j=0; j < old_size; j++) {

        Cor_SymTabElement *ste, *prev = NULL;
        for (ste = old_symtab[j]; ste; ste = ste->_next) {
     
            if (prev) {
     
                hval = prev->arg.symhash();
                indx = hval % _size;

                prev->set_next(_symtab[indx]);
                _symtab[indx] = prev;
            }

            prev = ste;
        }

        if (prev) {
     
            hval = prev->arg.symhash();
            indx = hval % _size;

            prev->set_next(_symtab[indx]);
            _symtab[indx] = prev;
        }
    }

    delete [] old_symtab;
}

// *************************************************************************
// For scanning the entire table
Cor_SymCursor *Cor_SymTable::OpenScan()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!this)
        return(NULL);

    Cor_SymCursor *cursor = new Cor_SymCursor(this);

    return(cursor);
}

// *************************************************************************
